/*
Ronald Speelman 2012-05-09
More documentation on: http://moinne.com/blog/ronald

 * **************************************************************
 * Examples
 * **************************************************************
 * SETUP A COUNTER:
 * Setup counter 'test1' with 5 rows and initial value 0 (Default):
 * call counter_setup('test1','test1',NULL,NULL);
 *
 * Setup counter 'test2' with 50 rows and initial value 1000:
 * call counter_setup('test2','test2',50,1000);
 *
 * Setup counter 'test3' with 1 row and initial value 0:
 * call counter_setup('testr3','test3',1,NULL);
 *
 * SELECT * FROM counters;
 *
 * INCREMENT THE VALUE OF A COUNTER:
 * increment counter 'test2':
 * call counter_increment('test2');
 *
 * GET THE VALUE OF A COUNTER:
 * get the value of counter 'test2':
 * SELECT counter_get('test2') AS total;
 *
 * RESET A COUNTER:
 * reset counter 'test2', set initial value to 9999 and update
 * the start_date to the current date
 * call counter_reset('test2',9999,TRUE);
 * SELECT * FROM counters WHERE counter_name = 'test2';
 *
 * DROP A COUNTER:
 * Drop counter 'test1'
 * call counter_drop('test1');
 */


/* set sql_mode to strict to avoid invalid variable assignments */
SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='STRICT_TRANS_TABLES';


/*****************************************************************
 * counter_setup
 *****************************************************************/
DELIMITER //
DROP PROCEDURE IF EXISTS counter_setup;
//
CREATE DEFINER = CURRENT_USER PROCEDURE counter_setup(
     IN p_name VARCHAR(50)
    ,IN p_description VARCHAR(500)
    ,IN p_counter_rows TINYINT
    ,IN p_init BIGINT(20)
    )
/**
 * Part of the counter functionality package. This package sets and maintains
 * counters. There are multiple rows for each named counter to avoid rowlocks in
 * the counter table when updating under heavy load.
 * <br>
 * %author Ronald Speelman
 * %version 1.2
 * First setup of a counter in the counter table.
 * %param p_name             String(50):  Name of the counter. e.g. 'homepage visits'
 * %param p_description      String(500): Description of the counter. e.g. 'log each homepage visits'
 * %param p_counter_rows     Number(255): Number of rows created for the counter. The more
 *                                        rows you create the less chance of a rowlock.
 *                                        Default = 5 rows for a counter
 * %param p_init             Number:      The initial value of the counter. Default = 0
 * %return No return value
*/
proc: BEGIN

        -- Declare variables
        DECLARE e_name_already_exists CONDITION FOR SQLSTATE '45000';
        DECLARE v_found_name INT DEFAULT 0;
        DECLARE v_counter_rows TINYINT DEFAULT 5;
        DECLARE v_init BIGINT DEFAULT 0;
        DECLARE v_iterate TINYINT DEFAULT 0;

        -- Use default values if NULL is given as parameter
        SET v_counter_rows = COALESCE(p_counter_rows, v_counter_rows);
        SET v_init = COALESCE(p_init, v_init);

        -- Exit if the counter_name already exists
        SELECT COUNT(1) INTO v_found_name FROM counters WHERE counter_name = p_name;
        IF v_found_name <> 0 THEN
            SIGNAL e_name_already_exists
                SET MESSAGE_TEXT = 'Cannot insert counter because counter_name already exists';
            LEAVE proc;
        END IF;

        -- insert counter
        loop1: LOOP
            SET v_iterate = v_iterate + 1;
            INSERT INTO counters (counter_name, counter_id, description,start_date,counter)
            VALUES ( p_name , v_iterate , p_description , NOW() , v_init);
            IF v_iterate >= v_counter_rows THEN
                LEAVE loop1;
            END IF;
            -- only the first row should have the initial value if not zero
            IF v_iterate = 1 THEN
                SET v_init = 0;
            END IF;
        END LOOP loop1;

END //
DELIMITER ;


/*****************************************************************
 * counter_increment
 *****************************************************************/
DELIMITER //
DROP PROCEDURE IF EXISTS counter_increment;
//
CREATE DEFINER = CURRENT_USER PROCEDURE counter_increment( IN p_name VARCHAR(50))
/**
 * Part of the counter functionality package. This package sets and maintains
 * counters. There are multiple rows for each named counter to avoid rowlocks in
 * the counter table when updating under heavy load.
 * <br>
 * %author Ronald Speelman
 * %version 1.2
 * Increment the value of a counter in the counter table.
 * A random row of a named counter will be used to update the value by one.
 * %param p_name             String(50):  Name of the counter. e.g. 'homepage visits'
 * %return No return value
*/
proc: BEGIN

        -- Declare variables
        DECLARE e_name_not_exists CONDITION FOR SQLSTATE '45000';
        DECLARE v_count_rows INT DEFAULT 0;
        DECLARE v_random_row TINYINT DEFAULT 1;

        -- find out how many row there are defined for this counter
        SELECT COUNT(1) INTO v_count_rows FROM counters WHERE counter_name = p_name;

        IF v_count_rows > 0 THEN -- update the counter

            -- pick a random row for the counter if we have more then one row
            IF v_count_rows > 1 THEN
                SET v_random_row =  FLOOR( RAND() * v_count_rows ) + 1;
            END IF;

            -- update the value
            UPDATE counters
                SET counter = counter + 1
            WHERE counter_name = p_name
                AND counter_id   = v_random_row;

        ELSE -- Exit if the counter_name not exists

            SIGNAL e_name_not_exists
                SET MESSAGE_TEXT = 'Cannot update the counter because counter_name does not exists';
            LEAVE proc;

        END IF;

END //
DELIMITER ;


/*****************************************************************
 * counter_get
 *****************************************************************/
DELIMITER //
DROP FUNCTION IF EXISTS counter_get;
//
CREATE FUNCTION counter_get( p_name VARCHAR(50))
    RETURNS BIGINT(20)
/**
 * Part of the counter functionality package. This package sets and maintains
 * counters. There are multiple rows for each named counter to avoid rowlocks in
 * the counter table when updating under heavy load.
 * <br>
 * %author Ronald Speelman
 * %version 1.2
 * Get the value of a counter in the counter table.
 * %param p_name             String(50):  Name of the counter. e.g. 'homepage visits'
 * %return total             Numver    :  The value of the counter
*/
proc: BEGIN

        -- Declare variables
        DECLARE e_name_not_exists CONDITION FOR SQLSTATE '45000';
        DECLARE v_count_rows INT DEFAULT 0;
        DECLARE v_total BIGINT DEFAULT NULL;

        -- get the value
        SELECT SUM(counter) INTO v_total
          FROM counters
         WHERE counter_name = p_name;

        -- Exit if the counter_name not exists
        IF v_total IS NULL THEN
            SIGNAL e_name_not_exists
                SET MESSAGE_TEXT = 'Cannot get the value of the counter because it does not exists';
            LEAVE proc;
        END IF;

        RETURN v_total;

END //
DELIMITER ;


/*****************************************************************
 * counter_reset
 *****************************************************************/
DELIMITER //
DROP PROCEDURE IF EXISTS counter_reset;
//
CREATE DEFINER = CURRENT_USER PROCEDURE counter_reset(
     IN p_name VARCHAR(50)
    ,IN p_init BIGINT(20)
    ,IN p_reset_startdate BOOLEAN
    )
/**
 * Part of the counter functionality package. This package sets and maintains
 * counters. There are multiple rows for each named counter to avoid rowlocks in
 * the counter table when updating under heavy load.
 * <br>
 * %author Ronald Speelman
 * %version 1.2
 * Reset the value of a counter in the counter table.
 * A random row of a named counter will be used to update the value by one.
 * %param p_name             String(50): Name of the counter. e.g. 'homepage visits'
 * %param p_init             Number    : The initial value of the counter
 * %param p_reset_startdate  Boolean   : Do you want to reset the startdate to
 *                                       the current date? Default = TRUE
 * %return No return value
*/
proc: BEGIN

        -- declare variables
        DECLARE e_name_not_exists CONDITION FOR SQLSTATE '45000';
        DECLARE v_found_name INT DEFAULT 0;
        DECLARE v_counter_rows TINYINT DEFAULT 1;
        DECLARE v_init BIGINT DEFAULT 0;
        DECLARE v_reset_startdate BOOLEAN DEFAULT TRUE;

        -- Use default values if NULL is given as parameter
        SET v_init = COALESCE(p_init, v_init);
        SET v_reset_startdate = COALESCE(p_reset_startdate, v_reset_startdate);

        -- Exit if the counter_name does not exists
        SELECT COUNT(1) INTO v_found_name FROM counters WHERE counter_name = p_name;
        IF v_found_name = 0 THEN
            SIGNAL e_name_not_exists
                SET MESSAGE_TEXT = 'Cannot reset counter because counter_name does not exists';
            LEAVE proc;
        END IF;

        -- reset counter
        UPDATE counters SET counter = v_init WHERE counter_name = p_name;
        IF v_init <> 0 THEN
            UPDATE counters
               SET counter = 0
             WHERE counter_name = p_name
               AND counter_id   > 1;
        END IF;

        -- reset start_date
        IF v_reset_startdate THEN
            UPDATE counters SET start_date = NOW() WHERE counter_name = p_name;
        END IF;


END //
DELIMITER ;


/*****************************************************************
 * counter_delete
 *****************************************************************/
DELIMITER //
DROP PROCEDURE IF EXISTS counter_delete;
//
CREATE DEFINER = CURRENT_USER PROCEDURE counter_delete( IN p_name VARCHAR(50))
/**
 * Part of the counter functionality package. This package sets and maintains
 * counters. There are multiple rows for each named counter to avoid rowlocks in
 * the counter table when updating under heavy load.
 * <br>
 * %author Ronald Speelman
 * %version 1.2
 * Delete a counter in the counter table.
 * %param p_name             String(50):  Name of the counter. e.g. 'homepage visits'
 * %return No return value
*/
proc: BEGIN

        -- Declare variables
        DECLARE e_name_not_exists CONDITION FOR SQLSTATE '45000';
        DECLARE v_found_name INT DEFAULT 0;

        -- Exit if the counter_name not exists
        SELECT COUNT(1) INTO v_found_name FROM counters WHERE counter_name = p_name;
        IF v_found_name <> 0 THEN
            DELETE FROM counters WHERE counter_name = p_name;
        ELSE
            SIGNAL e_name_not_exists
                SET MESSAGE_TEXT = 'Cannot delete counter because counter_name does not exists';
            LEAVE proc;
        END IF;

END //
DELIMITER ;


/* restore the original sql_mode */
SET sql_mode=@OLD_SQL_MODE;


